from . import visualize
from typing import Callable
import sys

TAPE_SIZE = 30_000


def get_matching_brackets(program: str) -> tuple[dict[int, int], dict[int, int]]:
	"""Gets the mappings required to match opening brackets to their closings, and vice-versa.

	Returns:
		A tuple containing two elements, the first of which mapping the opening brackets to their closing brackets. The
		second element is the opposite, mapping the closing brackets to their respective opening brackets."""
	opening_to_closing, closing_to_opening = {}, {}
	opened_brackets = []
	for pointer, command in enumerate(program):
		if command == "[":
			opened_brackets.append(pointer)
		if command == "]":
			matching_opening = opened_brackets.pop(-1)
			opening_to_closing[matching_opening] = pointer
			closing_to_opening[pointer] = matching_opening

	return opening_to_closing, closing_to_opening


def strip_bad_characters(program: str) -> str:
	"""Removes any characters that are not evaluate_brainfuck commands, ensuring comments and whitespace are allowed."""
	return "".join(filter(lambda x: x in [">", "<", "+", "-", ".", ",", "[", "]"], program))


def evaluate_processed(program: str,
					   input_callback: Callable[[], int] = None,
					   output_callback: Callable[[int], None] = None,
					   do_visualization: bool = False):
	"""Interprets `program`. Note that `program` should only contain evaluate_brainfuck commands, if any other characters are
	passed, it will likely behave fine, but undefined.

	Args:
		program: The program to run.
		input_callback: The function that is called when the program requests input (i.e. encounters a `,`). This should
			return an integer to write to the cell. Will default to prompting the user for input if left blank.
		output_callback: The function that is called when the program wants to output a byte. Should take an integer as
			an argument. Will default to writing to stdout if left blank.
		do_visualization: Whether the function should display its operations. Defaults to `False`."""
	if input_callback is None:
		input_callback = lambda: int(input())
	if output_callback is None:
		output_callback = lambda x: sys.stdout.write(chr(x))
	tape = [0 for _ in range(TAPE_SIZE)]
	head_position = 0
	code_position = 0

	opening_to_closing, closing_to_opening = get_matching_brackets(program)

	while code_position < len(program):
		cmd = program[code_position]

		if cmd == ">":
			head_position = (head_position + 1) % TAPE_SIZE
		elif cmd == "<":
			head_position = (head_position - 1) % TAPE_SIZE
		elif cmd == "+":
			tape[head_position] += 1
		elif cmd == "-":
			tape[head_position] -= 1
		elif cmd == ".":
			output_callback(tape[head_position])
		elif cmd == ",":
			tape[head_position] = input_callback()

		if cmd == "[" and tape[head_position] == 0:
			code_position = opening_to_closing[code_position]
		elif cmd == "]" and tape[head_position] != 0:
			code_position = closing_to_opening[code_position]

		if do_visualization:
			visualize.write_visualization(program, tape, code_position, head_position)
		code_position += 1


def evaluate_brainfuck(program: str, **kwargs):
	"""Completely evaluates and runs `program`.

	Args:
		program: The program to run.
		*args: For usage please see `evaluate_brainfuck.evaluate_processed`."""
	program = strip_bad_characters(program)
	evaluate_processed(program, **kwargs)



from .brainfuck import evaluate_brainfuck
from .visualize import visualize_evaluation, set_visualization_refresh_rate
import sys, os, select


def incorrect_usage():
	sys.stdout.write("Incorrect usage: use the -h flag to display a help message. \n")
	sys.exit(1)


def help_message():
	sys.stdout.write("""NAME	
	brainfuck.py -- A lightweight pure python brainfuck interpreter

SYNOPSIS
	brainfuckpy [-h] [-vis [-vrr rate]] [file|program]

DESCRIPTION
	Either pipe in, provide as an argument or provide a file containing a valid program.
	
	-vis --visualize
			Run the code visualizer. Optionally the -vrr flag can be used to set a refresh rate of the visualization.
	-vrr --visualization-refresh-rate
			Set the amount of times per second the visualizer updates. Provide this refresh rate as the immediate next
			argument.
	-h, --help
			Display this message.\n""")


def main():
	# Please ignore this entire first block, it is distinctly awful
	evaluator = evaluate_brainfuck
	if "-vis" in sys.argv or "--visualize" in sys.argv:
		evaluator = visualize_evaluation
		try:
			if "-vrr" in sys.argv:
				speed_index = sys.argv.index("-vrr") + 1
				set_visualization_refresh_rate(float(sys.argv[speed_index]))
			elif "--visualization-refresh-rate" in sys.argv:
				speed_index = sys.argv.index("--visualization-refresh-rate") + 1
				set_visualization_refresh_rate(float(sys.argv[speed_index]))
		except ValueError:
			incorrect_usage()
	elif "-vrr" in sys.argv or "--visualization-refresh-rate" in sys.argv:
		incorrect_usage()

	if len(sys.argv) == 1 and select.select([sys.stdin, ], [], [], 0.0)[0]:
		prgm = sys.stdin.read()
		evaluator(prgm)
	elif len(sys.argv) == 1:
		incorrect_usage()
	elif os.path.isfile(sys.argv[-1]):
		with open(sys.argv[-1]) as file:
			prgm = file.read()
		evaluator(prgm)
	elif sys.argv[1] in {"-h", "--help"}:
		help_message()
	elif sys.argv[-1]:
		prgm = sys.argv[-1]
		evaluator(prgm)
	else:
		incorrect_usage()


if __name__ == '__main__':
	main()



from . import brainfuck
from rich.console import Console
import time

console = Console()

CODE_ACCENT = "bold red"
TAPE_ACCENT = "bold red"

seconds_per_refresh = 0.1

outputs = []


def _visualization_output_callback(byte: int):
	outputs.append(chr(byte))


def _visualization_input_callback() -> int:
	return int(console.input("\n"))


def set_visualization_refresh_rate(speed: float):
	global seconds_per_refresh
	seconds_per_refresh = speed


def get_subset(tape: list[str], centered_at: int, length: int=40) -> list[str]:
	"""Gets a subset of the tape of length `length` in order to make the visualizer wrap lines better."""
	minimum = centered_at - length//2
	maximum = centered_at + length//2

	if minimum < 0:
		return tape[minimum:] + tape[:maximum]
	elif maximum > brainfuck.TAPE_SIZE:
		maximum %= brainfuck.TAPE_SIZE
		return tape[minimum:] + tape[:maximum]
	else:
		return tape[minimum:maximum]


def write_visualization(program: str, tape: list[int], code_pointer: int, head_position: int):
	"""Writes a single cycle of the visualization, means clearing and displaying the information."""
	# construct the tape
	tape_with_highlight = [i for i in tape]
	tape_with_highlight[head_position] = f"[{TAPE_ACCENT}]{tape[head_position]}[/{TAPE_ACCENT}]"
	tape_with_highlight = get_subset(tape_with_highlight, head_position)
	tape_text = ", ".join(str(i) for i in tape_with_highlight)

	# construct the program
	program_text = program[:code_pointer] + f"[{CODE_ACCENT}]" + program[code_pointer] + f"[/{CODE_ACCENT}]" + program[code_pointer+1:]

	# construct the outputs
	output_text = "".join(outputs)

	console.clear()
	console.print(tape_text)
	console.print("\n")
	console.print(program_text)
	console.print("\n")
	console.print(output_text)

	time.sleep(seconds_per_refresh)


def visualize_evaluation(program: str):
	"""Completely visualizes the execution of `program`."""
	brainfuck.evaluate_brainfuck(
		program,
		input_callback=_visualization_input_callback,
		output_callback=_visualization_output_callback,
		do_visualization=True
	)

